local chess = {}

chess.moves = require("chess.moves")
chess.pgn = require("chess.pgn")
chess.fen = require("chess.fen")

local char = string.char
local floor = math.floor

local serialize = require("chess.serialize")

-- 1  2  3  4  5  6  7  8
-- 9  10 11 12 13 14 15 16
-- 17 18 19 20 21 22 23 24
-- 25 26 27 28 29 30 31 32
-- 33 34 35 36 37 38 39 40
-- 41 42 43 44 45 46 47 48
-- 49 50 51 52 53 54 55 56
-- 57 58 59 60 61 62 63 64

function chess.new(fen)
  fen = fen or "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 0"
  return chess.fen.board(fen)
end

function chess.getMoves(b)
  local moves = {}
  -- for each piece
  chess.moves.getAll(b, moves)
  assert(#moves%2 == 0, 'odd number of move pairs')
  return moves
end

function chess.move(b, from, to, promo)
  assert(from >= 1 and from <= 64, "invalid from square")
  assert(to >= 1 and to <= 64, "invalid to square")
  assert(to ~= from, "move squares are the same")
  
  local p1 = b[from]
  local p2 = b[to]
  local p3 = p1
  
  assert(p1, "starting square is empty")
  assert(p2 == nil or p2 <= 10, "cannot capture king")

  local csq = to
  -- pawn move
  if p1 <= 2 then
    -- en passant capture
    if to == b.eps then
      local advance = 8*((b.stm + 1)%2*2 - 1)
      csq = b.eps + advance -- todo
    end
    -- promotion
    if to <= 8 or to >= 57 then
      -- assume we promote to queen
      p3 = promo or (p1 + 8)
    end
  end
  
  b[from] = nil
  b[csq] = nil
  b[to] = p3

  -- Side to move
  local stm1 = b.stm
  local stm2 = (stm1 + 1)%2
  b.stm = stm2
  -- Castling availability
  local csa = b.csa
  local ws = csa%4
  local bs = (csa - ws)/4
  if p1 >= 11 then
    ws = ws*stm2
    bs = bs*stm1
  elseif p1 == 7 or p1 == 8 then
    -- todo
  end
  b.csa = ws + bs*4
  -- En passant square
  b.eps = 0
  if p1 <= 2 then
    local offset = to - from
    if offset == 16 and from >= 9 and from <= 16 then
      b.eps = from + 8
    elseif offset == -16 and from >= 49 and from <= 56 then
      b.eps = from - 8
    end
  end
  -- Half move clock
  b.hmc = (p2 or p1 <= 2) and 0 or (b.hmc + 1)
  -- Ply
  b.ply = b.ply + 1
end

local b0 = string.byte('1') - 1
local bA = string.byte('a') - 1

function chess.toASCII(b)
  local sz = {}
  sz[#sz + 1] = " abcdefgh \n"
  for y = 8, 1, -1 do
    sz[#sz + 1] = string.char(y + b0)
    for x = 1, 8 do
      local i = x + (y - 1)*8
      local p = b[i]
      sz[#sz + 1] = serialize.piece.str[p or 0]
      if i%8 == 0 then
        sz[#sz + 1] = "\n"
      end
    end
  end
  return table.concat(sz)
end

function chess.toAlgebraic(b, from, to, p3)
  return chess.pgn.algebraic(b, from, to, p3)
end

function chess.toLAN(b, from, to, p3)
  return chess.pgn.lan(b, from, to, p3)
end

function chess.toSAN(b, from, to, p3)
  local l, f, r = chess.moves.legal(b, from, to)
  return chess.pgn.san(b, from, to, p3, f, r)
end

function chess.fromSAN(b, san)
  local moves = chess.getMoves(b) --chess.available(b)
  for i = 1, #moves, 2 do
    local from = moves[i]
    local to = moves[i + 1]
    local san2 = chess.toSAN(b, from, to)
    if san == san2 then
      return from, to
    end
  end
end

return chess