local fen = {}

local len = string.len
local byte = string.byte
local sub = string.sub
local gmatch = string.gmatch
local char = string.char
local concat = table.concat

local serialize = require("chess.serialize")

--- Parses ForsythEdwards Notation
-- @param fen FEN notation
-- @return table
function fen.board(fen)
  -- break string into sections
  if sub(fen, -1) ~= " " then
    fen = fen .. " "
  end
  local s = {}
  for sz in gmatch(fen, "(.-) ") do
    s[#s + 1] = sz
  end
  local board = {}
  -- 1.Piece placement
  local s1 = s[1]
  if s1 then
    local x = 1
    local y = 8
    for i = 1, len(s1) do
      local b = byte(s1, i)
      if b >= 49 and b <= 56 then -- 1-8
        -- empty square
        x = x + (b - 48)
      elseif b == 47 then -- /
        -- rank
        x = 1
        y = y - 1
      elseif b == 32 then -- space
        -- end of section
        break
      else
        -- piece
        local c = char(b)
        --if not white[c] and not black[c] then
          --error("invalid symbol:" .. c)
        --end
        local sq = x + (y - 1)*8
        board[sq] = serialize.piece.bin[c or 0]
        x = x + 1
      end
    end
  end
  -- 2.Side to move
  board.stm = serialize.stm.bin[s[2] or 0]
  -- 3.Castling ability
  board.csa = serialize.csa.bin[s[3] or 0]
  -- 4.En Passant square
  board.eps = serialize.sq.bin[s[4] or 0]
  -- 5.Halfmoves
  local s5 = s[5]
  local hmc = tonumber(s[5] or 0) or 0
  if hmc < 0 then
    hmc = -hmc
  end
  board.hmc = hmc
  -- 6.Fullmove counter
  local fmc = tonumber(s[6] or 0) or 0
  if fmc < 0 then
    fmc = -fmc
  end
  fmc = fmc*2
  if board.stm == serialize.stm.bin.b then
    fmc = fmc + 1
  end
  board.ply = fmc
  return board
end

function fen.string(board)
  local fen = {}
  -- 1.Piece placement
  for y = 8, 1, -1 do
    for x = 1, 8 do
      local i = x + (y - 1)*8
      local p = board[i]
      if p then
        fen[#fen + 1] = serialize.piece.str[p]
      else
        local n = fen[#fen]
        if type(n) == "number" then
          fen[#fen] = n + 1
        else
          fen[#fen + 1] = 1
        end
      end
    end
    if y > 1 then
      fen[#fen + 1] = "/"
    end
  end
  -- 2.Side to move
  fen[#fen + 1] = " " .. serialize.stm.str[board.stm]
  -- 3.Castling
  fen[#fen + 1] = " " .. serialize.csa.str[board.csa]
  -- 4.En Passant
  fen[#fen + 1] = " " .. serialize.sq.str[board.eps]
  -- 5.Halfmoves
  fen[#fen + 1] = " " .. board.hmc
  -- 6.Fullmove count
  local fmc = board.ply
  if board.stm == serialize.stm.bin.b then
    fmc = fmc + 1
  end
  fmc = fmc/2
  fen[#fen + 1] = " " .. fmc
  return concat(fen)
end

return fen