local tremove = table.remove

-- initial position
local inital64 =
{
  7,  3,  5,  9, 11,  5,  3,  7,
  1,  1,  1,  1,  1,  1,  1,  1,
  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,
  2,  2,  2,  2,  2,  2,  2,  2,
  8,  4,  6, 10, 12,  6,  4,  8
}

-- 10x12 indexing
local mailbox120 =
{
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1,  1,  2,  3,  4,  5,  6,  7,  8, -1,
  -1,  9, 10, 11, 12, 13, 14, 15, 16, -1,
  -1, 17, 18, 19, 20, 21, 22, 23, 24, -1,
  -1, 25, 26, 27, 28, 29, 30, 31, 32, -1,
  -1, 33, 34, 35, 36, 37, 38, 39, 40, -1,
  -1, 41, 42, 43, 44, 45, 46, 47, 48, -1,
  -1, 49, 50, 51, 52, 53, 54, 55, 56, -1,
  -1, 57, 58, 59, 60, 61, 62, 63, 64, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1
}

-- 8x8 indexing
local mailbox64 =
{
  22, 23, 24, 25, 26, 27, 28, 29,
  32, 33, 34, 35, 36, 37, 38, 39,
  42, 43, 44, 45, 46, 47, 48, 49,
  52, 53, 54, 55, 56, 57, 58, 59,
  62, 63, 64, 65, 66, 67, 68, 69,
  72, 73, 74, 75, 76, 77, 78, 79,
  82, 83, 84, 85, 86, 87, 88, 89,
  92, 93, 94, 95, 96, 97, 98, 99
}

-- attack offsets
local attacks120 =
{
  { 9, 11 },
  { -11, -9 },
  { -21, -19, -12, -8, 8, 12, 19, 21 },
  { -21, -19, -12, -8, 8, 12, 19, 21 },
  { -11, -9, 9, 11 },
  { -11, -9, 9, 11 },
  { -10, -1, 1, 10 },
  { -10, -1, 1, 10 },
  { -11, -10, -9, -1, 1, 9, 10, 11 },
  { -11, -10, -9, -1, 1, 9, 10, 11 },
  { -11, -10, -9, -1, 1, 9, 10, 11 },
  { -11, -10, -9, -1, 1, 9, 10, 11 }
}

-- advance offsets
local advances64 = { 8, -8 }

local sliding = { 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0 }

local rays = { 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 0, 0 }

local rays2 =
{
  [0] = { 0, 0, 0, 0, 0, 1, 0, 1, 0, 2, 0, 0 },
  [1] = { 0, 0, 0, 0, 1, 0, 1, 0, 2, 0, 0, 0 }
}

-- current board
local b
-- generated moves
local m
-- number of generated moves
local n = 0
-- side to move
local side1, side2
-- king and en passant squares, castling availability
local ksq, eps, csa

--- Helper functions

--- Finds the first occurrence of a piece
-- @param p1 Piece
-- @return Square number or nil
local function findPiece(p1)
  for from = 1, 64 do
    if p1 == b[from] then
      return from
    end
  end
end

local function blocked(from120, to120, offset120)
  local p2
  repeat
    from120 = from120 - offset120
    local from = mailbox120[from120]
    -- off the board
    if from == -1 then
      break
    end
    p2 = b[from]
  until p2
  return p2
end

--- Checks if a square is attacked
-- @param stm Side to move
-- @param to Square number between 1 and 64
-- @return True if attacked by the moving side
local function attacked(stm, to)
  local to120 = mailbox64[to]
  --assert(to120, "invalid to square")
  -- for each piece type
  for p1 = 2 - stm, 12, 2 do
    local offsets = attacks120[p1]
    --assert(offsets, "invalid piece on the board")
    -- for each offset
    local stop = rays[p1] == 0
    for i = 1, #offsets do
      local offset120 = offsets[i]
      local from120 = to120
      repeat
        from120 = from120 - offset120
        local from = mailbox120[from120]
        -- off the board
        if from == -1 then
          break
        end
        local p2 = b[from]
        if p2 == p1 then
          return true
        end
        -- blocked
      until p2 or stop
    end
  end
  return false
end

--- Move generation functions

--- Prepares move generation
-- @param board Board table
-- @param moves Moves list
local function genInit(board, moves)
  b = board
  m = moves or {}
  n = #m + 1
  side1 = board.stm
  side2 = (side1 + 1)%2
  ksq = board.ksq or findPiece(12 - side1)
  eps = board.eps
  csa = board.csa
end

--- Generates one move
-- @param from Starting square
-- @param to Ending square
-- @param p1 Piece on the starting square
-- @param p2 Piece on the ending square
local function genMove(from, to, p1, p2)
  --assert(b[from] == p1)
  --assert(b[to] == p2)
  local ksq2 = ksq
  if ksq == from then
    ksq2 = to
  end
  b[from], b[to] = nil, p1
  if not attacked(side2, ksq2) then
    m[n], m[n + 1] = from, to
    n = n + 2
  end
  b[from], b[to] = p1, p2
end

-- Generates moves to a given square
-- @param from Ending square
-- @param p1 Piece type or 0 for any
local function genMovesTo(to, _p1)
  local to120 = mailbox64[to]
  --assert(to120, "invalid to square")
  local p, k = _p1, _p1
  if _p1 == 0 then
    p, k = 2 - side1, 12 - side1
  end
  assert(p%2 == side1)
  local p2 = b[to]
  -- attacking own piece
  if p2 and p2%2 == side1 then
    return
  end
  -- pawn advance moves
  local advance = advances64[2 - side1]
  if p <= 2 and p2 == nil then
    --assert(from > 8 and from <= 56, "invalid pawn square")
    local from1 = to - advance
    if b[from1] == p then
      -- advance one
      genMove(from1, to, p)
    elseif b[from1] == nil then
      local from2 = from1 - advance
      if p == inital64[from2] then
        if b[from2] == p then
          -- advance two
          genMove(from2, to, p)
        end
      end
    end
  end
  -- attack moves
  for p1 = p, k, 2 do
    local offsets = attacks120[p1]
    --assert(offsets, "invalid piece on the board")
    -- for each offset
    local stop = sliding[p1] == 0
    for i = 1, #offsets do
      local offset120 = offsets[i]
      local from120 = to120
      repeat
        from120 = from120 - offset120
        local from = mailbox120[from120]
        -- off the board
        if from == -1 then
          break
        end
        local p1b = b[from]
        if p1b then
          -- capture
          if p1b == p1 then
            if p1 <= 2 and p2 == nil then
              if to == eps then
                -- en passant capture
                local csq = eps + advance/2
                local p3 = b[csq]
                b[csq] = nil
                genMove(from, to, p1)
                b[csq] = p3
              end
              break
            else
              genMove(from, to, p1, p2)
            end
          end
          -- blocked
          break
        end
      until stop
    end
  end
end

-- Generates moves for one piece
-- @param from Starting square
-- @param p1 Piece on the starting square
local function genMovesFrom(from, p1)
  local advance = advances64[2 - side1]
  local from120 = mailbox64[from]
  -- pawn advance moves
  if p1 <= 2 then
    --assert(from > 8 and from <= 56, "invalid pawn square")
    local to1 = from + advance
    if b[to1] == nil then
      -- advance one
      genMove(from, to1, p1)
      if p1 == inital64[from] then
        local to2 = to1 + advance
        if b[to2] == nil then
          -- advance two
          genMove(from, to2, p1)
        end
      end
    end
  end
  -- attack moves
  local stop = sliding[p1] == 0
  local offsets = attacks120[p1]
  --assert(offsets, "invalid piece on the board")
  -- for each offset
  for i = 1, #offsets do
    local to120 = from120
    local offset120 = offsets[i]
    repeat
      to120 = to120 + offset120
      local to = mailbox120[to120]
      -- off the board
      if to == -1 then
        break
      end
      local p2 = b[to]
      if p2 then
        -- own piece
        if side1 ~= p2%2 then
          -- capture
          genMove(from, to, p1, p2)
        end
        break
      else
        if p1 <= 2 then
          -- pawn
          if to == eps then
            -- en passant capture
            local csq = eps + advance/2
            local p3 = b[csq]
            b[csq] = nil
            genMove(from, to, p1)
            b[csq] = p3
          end
          break
        else
          -- quiet move
          genMove(from, to, p1)
        end
      end
    until stop
  end
end

local function genMovesCastle()
  -- castling
  -- king in place
  local k = b[ksq]
  if k == inital64[ksq] then
    -- not in check
    if not attacked(side2, ksq) then
      -- king side
      local rksq = ksq + 3
      if b[rksq] == inital64[rksq] then
        -- blocked
        if not (b[ksq + 1] or b[ksq + 2]) then
          -- passes through check
          if not attacked(side2, ksq + 1) then
            genMove(ksq, ksq + 2, k)
          end
        end
      end
      -- queen side
      local rqsq = ksq - 4
      if b[rqsq] == inital64[rqsq] then
        -- blocked
        if not (b[ksq - 1] or b[ksq - 2] or b[ksq - 3]) then
          -- passes through check
          if not attacked(side2, ksq - 1) then
            genMove(ksq, ksq - 2, k)
          end
        end
      end
    end
  end
end

--- Generates all moves for a given side
-- @return Moves list
local function genMoves()
  -- for each piece on the board
  for from = 1, 64 do
    local p1 = b[from]
    if p1 and p1%2 == side1 then
      genMovesFrom(from, p1)
    end
  end
end

local gen = {}

--- Get all moves to a given square
-- @param board Board table
-- @param to Square number
-- @param p1 Piece (optional)
-- @param moves Moves table (optional)
-- @return Moves table
function gen.getTo(board, to, p1, moves)
  if p1 and p1 > 0 and p1%2 ~= board.stm then
    return
  end
  genInit(board, moves)
  genMovesTo(to, p1 or 0)
  return m
end

--- Get all moves from a given square
-- @param board Board table
-- @param from Square number
-- @param moves Moves table
-- @return Moves table
function gen.getFrom(board, from, moves)
  genInit(board, moves)
  local p1 = board[from]
  if p1 == nil or p1 == 0 or p1%2 ~= side1 then
    return
  end
  genMovesFrom(from, p1)
  if p1 >= 11 then
    genMovesCastle()
  end
  return m
end

--- Get all moves
-- @param board Board table
-- @param stm Side to move
-- @param eps En Passant square
-- @param csa Castling availability
-- @param moves Moves table
-- @return Moves table
function gen.getAll(board, moves)
  genInit(board, moves)
  genMoves()
  genMovesCastle()
  return m
end

function gen.check(board)
  genInit(board)
  return attacked(side2, ksq)
end

function gen.checkmate(board)
  genInit(board)
  if not attacked(side2, ksq) then
    return false
  end
  genMoves()
  genMovesCastle()
  return #m == 0
end

-- @param return legal, file, rank
function gen.legal(board, from, to)
  local p1 = board[from]
  if p1 and p1 > 0 and p1%2 ~= board.stm then
    -- no piece on source square
    -- wrong side to move
    return false, false, false
  end
  genInit(board)
  genMovesTo(to, p1)
  for i = 1, #m, 2 do
    if m[i] == from and m[i + 1] == to then
      return true
    end
  end
  return false
  --[[
  if #m == 0 then
    -- illegal
    return false, false, false
  elseif #m == 2 then
    -- legal and unambiguous
    return true, false, false
  end
  if p1 >= 11 then
    return true, false, false
  elseif p1 <= 2 then
    -- file for pawns
    return true, true, false
  end
  -- disambiguation
  local files = false
  local ranks = false
  for i = 1, #m, 2 do
    local from2 = m[i]
    if from ~= from2 then
      local offset = from2 - from
      if offset < 0 then
        offset = -offset
      end
      local ox = offset%8
      local oy = (offset - ox)/8
      files = files or (ox == 0)
      ranks = ranks or (oy == 0)
      if files and ranks then
        break
      end
    end
  end
  if not files then
    -- 1.file of departure if different
    return true, true, false
  elseif not ranks then
    -- 2.rank of departure if the files are the same
    -- but the ranks differ
    return true, false, true
  else
    -- 3.the complete origin square coordinate otherwise
    return true, true, true
  end
  ]]
end

return gen