local pgn = {}

local match = string.match
local gmatch = string.gmatch
local sub = string.sub
local len = string.len
local byte = string.byte
local find = string.find
local upper = string.upper
local format = string.format

function pgn.moves(s, i)
  local moves = {}
  moves.Annotations = {}
  local i = i or 1
  while i <= len(s) do
    local b = byte(s, i)
    if b == 40 then -- (
      -- start of variation
      i = i + 1
      moves[#moves + 1], i = pgn.moves(s, i)
    elseif b == 41 then -- )
      -- end of variation
      i = i + 1
      return moves, i
    elseif b == 123 then -- {
      -- annotation
      local j = find(s, "}", i + 1, true)
      assert(j, "unfinished annotation")
      moves.Annotations[#moves] = sub(s, i + 1, j - 1)
      i = j
    elseif b >= 49 and b <= 57 then -- 1-9
      -- move number indication
      local n2, ws = match(s, "^(%d+)[%.]*(%s*)", i)
      i = i + len(n2) + len(ws or "")
    else
      -- SAN move
      local san = match(s, "^([%a][%w%+%=%#%-]+)", i)
      if san then
        i = i + len(san)
        moves[#moves + 1] = san
      else
        i = i + 1
      end
    end
  end
  return moves, i
end

-- PGN text lines are limited to having a maximum of 255 characters per line
-- including the newline character.
-- Lines with 80 or more printing characters are strongly discouraged
-- because of the difficulties experienced by common text editors with long lines.

--[[
Seven Tag Roster
================
1.Event: the name of the tournament or match event.
2.Site: the location of the event. This is in "City, Region COUNTRY" format, where COUNTRY is the three-letter International Olympic Committee code for the country. An example is "New York City, NY USA".
3.Date: the starting date of the game, in YYYY.MM.DD form. "??" are used for unknown values.
4.Round: the playing round ordinal of the game within the event.
5.White: the player of the white pieces, in "last name, first name" format.
6.Black: the player of the black pieces, same format as White.
7.Result: the result of the game. This can only have four possible values: "1-0" (White won), "0-1" (Black won), "1/2-1/2" (Draw), or "*" (other, e.g., the game is ongoing).
]]

--local str = { "Event", "Site", "Date", "Round", "White", "Black", "Result" }

-- scan PGN file without parsing the moves
function pgn.scan(s)
  local scan = {}
  local game = nil
  if sub(s, -1) ~= "\n" then
    s = s .. "\n"
  end
  for line in gmatch(s, "(.-)\n") do
    -- extract tag and value
    local tag, value = match(line, '%[(%w+) "(.*)"%]')
    if tag and value then
      if game == nil then
        game = {}
        scan[#scan + 1] = game
      end
      -- line contains tag pair
      game[tag] = value
    else
      -- line contains movetext
      scan[#scan].Movetext = (scan[#scan].Movetext or "") .. line .. "\n"
      game = nil
    end
  end
  return scan
end

local serialize = require("chess.serialize")

local squares = serialize.sq.str
local pieces = serialize.piece.str

--- Long algebraic notation
-- Nb1-c3
function pgn.lan(b, from, to, p3, p, s, c)
  local p1 = board[from]
  local p2 = board[to]
  -- Castling
  if p1 >= 11 then
    local offset = to - from
    local ox = offset%8
    if ox == -2 then
      return 'O-O-O'
    elseif ox == 2 then
      return 'O-O'
    end
  end
  -- 1.Piece type
  local s1 = p
  if s1 == nil then
    s1 = ''
    if p1 >= 3 then
      s1 = upper(pieces[p1])
    end
  end
  -- 2.Source
  local s2 = s
  if s2 == nil then
    s2 = squares[from]
  end
  -- 3.Capture
  local s3 = c
  if c == nil then
    s3 = '-'
    if p2 or (p1 <= 2 and to == b.eps) then
      s3 = 'x'
    end
  end
  -- 4.Destination
  local s4 = squares[to]
  -- 5.Promotion
  local s5 = ''
  if p1 <= 2 then
    if to <= 8 or to >= 56 then
      s5 = pieces[p3] or 'Q'
    end
  end
  return format("%s%s%s%s%s", s1, s2, s3, s4, s5)
end

--- Algebraic notation
-- e2e4
function pgn.algebraic(b, from, to, p3)
  return pgn.lan(b, from, to, p3, '', nil, '')
end

local pieces_iccn = { 0,0, 4,4, 3,3, 2,2, 1,1, 0,0 }
function pgn.iccn(b, from, to, p3)
  -- 1.From square
  local x1 = (from - 1)%8 + 1
  local y1 = (from - x1)/8 + 1
  -- 2.To square
  local x2 = (to - 1)%8 + 1
  local y2 = (to - x2)/8 + 1
  -- 3.Promotion code
  local z = ''
  if p3 then
    z = pieces_iccn[p3]
  end
  return format("%d%d%d%d%s", x1, y1, x2, y2, z)
end

--- Standard algebraic notation
-- e4
-- @param from From square
-- @param to To square
-- @param p3 Promotion piece
-- @param r Show source rank
-- @param f Show source file
function pgn.san(b, from, to, p3, f, r)
  local p1 = b[from]
  local p2 = b[to]
  local c = ''
  if p2 then
    c = 'x'
  end
  if p1 <= 2 then
    f, r = false, false
    if p2 or to == b.eps then
      f, r = true, false
      c = 'x'
    end
  end
  local s = ''
  if f and r then
    s = squares[from]
  elseif f then
    s = sub(squares[from], 1, 1)
  elseif r then
    s = sub(squares[from], 2, 2)
  end
  return pgn.lan(b, from, to, p3, nil, s, c)
end

return pgn